/*
 * Copyright (C) 2014 Alina Trifan and António Neves
 * mailto: alina.trifan@ua.pt -- an@ua.pt 
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "CameraFirewire.h"
#include "CameraSettings.h"
#include <errno.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <dc1394/format7.h>

namespace uav{

int CameraFirewire::initCamera(CameraSettings *camSettings)
{
	dc1394error_t err;
	dc1394camera_list_t *cameraList;
	dc1394_t *d;
	unsigned camCount;

	// ---------------------------------
	// find the camera and connect to it

	// get the list of connected cameras
	d = dc1394_new ();
	err=dc1394_camera_enumerate(d, &cameraList);
	DC1394_ERR_RTN(err,"Failed to enumerate cameras");

	if (cameraList->num == 0)
	{
		dc1394_log_error("No cameras found - trying a reset bus");
		err=dc1394_camera_reset(camera);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR camera reset.\n");
		err=dc1394_reset_bus(camera);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR bus reset.\n");

		err=dc1394_camera_enumerate(d, &cameraList);
		DC1394_ERR_RTN(err,"Failed to enumerate cameras");

		if (cameraList->num == 0)
		{
			dc1394_log_error("No cameras found - stopping");
			exit(1);
		}
	}


		camera = dc1394_camera_new(d, cameraList->ids[0].guid);
		std::cerr<<"Camera "<<camCount<<"has model name "<<camera->model<<std::endl;




	// connect to the camera
	if (camera == NULL)
	{
		dc1394_log_error("Failed to initialize camera with guid %llx", cameraList->ids[0].guid);
		exit(1);
	}
	dc1394_camera_free_list(cameraList);

	// ----------------------------------------------------
	// setup all the variables to proceed to the camera
	// configuration

	// convert FPS's

	switch(camSettings->getCameraSetting(UAV_FPS))
	{
	case 1:
		err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_1_875);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
		break;
	case 3:
		err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_3_75);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
		break;
	case 7:
		err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_7_5);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
		break;
	case 15:
		err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_15);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
		break;
	case 30:
		err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_30);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
		break;

	case 60:
		err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_60);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
		break;
	case 120:
		err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_120);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
		break;
	case 240:
		err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_240);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
		break;

	default:
		err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_MIN);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
		break;
	}

	// are we using the format 7?
	if(camSettings->getCameraSetting(UAV_FORMAT7) == 1)
	{
		unsigned maxCols, maxRows;

		// get maximum resolution available
		err=dc1394_format7_get_max_image_size(camera, DC1394_VIDEO_MODE_FORMAT7_0, &maxCols, &maxRows);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: get max image size error.\n");

		// try to set our image size... or the maximum available

		if(camSettings->getCameraSetting(UAV_NCOLS) >= maxCols)
			camSettings->setCameraSetting(UAV_NCOLS, maxCols);

		if(camSettings->getCameraSetting(UAV_NROWS) >= maxRows)
			camSettings->setCameraSetting(UAV_NROWS, maxRows);

		if(camSettings->getCameraSetting(UAV_CCDCOL) + camSettings->getCameraSetting(UAV_NCOLS) >= maxCols)
			camSettings->setCameraSetting(UAV_CCDCOL, 0);

		if(camSettings->getCameraSetting(UAV_CCDROW) + camSettings->getCameraSetting(UAV_NROWS) >= maxRows)
			camSettings->setCameraSetting(UAV_CCDROW, 0);

		err=dc1394_video_set_mode(camera, DC1394_VIDEO_MODE_FORMAT7_0);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set MODE.\n");



		// display some information about the choices made
		std::cerr << "Camera: cols x rows (" << camSettings->getCameraSetting(UAV_NCOLS) << "x" << camSettings->getCameraSetting(UAV_NROWS) << ")" <<
				", max(" << maxCols << "x" << maxRows << ")." << std::endl;
		std::cerr << "Camera: ccdC x ccdR (" << camSettings->getCameraSetting(UAV_CCDCOL) << "x" << camSettings->getCameraSetting(UAV_CCDROW) << ")." << std::endl;
	}


	// convert operation mode
	switch(camSettings->getCameraSetting(UAV_OPMODE))
	{
	case 0:
		err=dc1394_video_set_operation_mode(camera, DC1394_OPERATION_MODE_LEGACY );
		DC1394_ERR_CLN_RTN(err,shutDownCamera(), "\nCamera: ERROR set operation mode.\n");
		break;
	case 1:
		err=dc1394_video_set_operation_mode(camera, DC1394_OPERATION_MODE_1394B);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(), "\nCamera: ERROR set operation mode.\n");
		break;
	default:
		err=dc1394_video_set_operation_mode(camera, DC1394_OPERATION_MODE_MIN);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(), "\nCamera: ERROR set operation mode.\n");
		break;

	}
	// ISO speed equal/above 800 needs MODE-1394b otherwise use LEGACY
	if( camSettings->getCameraSetting(UAV_SPEED) >= 800)
	{
		err=dc1394_video_set_operation_mode(camera, DC1394_OPERATION_MODE_1394B);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(), "\nCamera: ERROR set operation mode.\n");
	}
	else
	{
		err=dc1394_video_set_operation_mode(camera, DC1394_OPERATION_MODE_LEGACY);

		DC1394_ERR_CLN_RTN(err,shutDownCamera(), "\nCamera: ERROR set operation mode.\n");
	}

	// convert speed
		switch(camSettings->getCameraSetting(UAV_SPEED))
		{
		case 100:
			err=dc1394_video_set_iso_speed(camera,DC1394_ISO_SPEED_100 );
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set ISO speed.\n");
			break;
		case 200:
			err=dc1394_video_set_iso_speed(camera,DC1394_ISO_SPEED_200 );
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set ISO speed.\n");
			break;
		case 400:
			err=dc1394_video_set_iso_speed(camera,DC1394_ISO_SPEED_400 );
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set ISO speed.\n");
			break;
		case 800:
			err=dc1394_video_set_iso_speed(camera,DC1394_ISO_SPEED_800 );
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set ISO speed.\n");
			break;
		case 1600:
			err=dc1394_video_set_iso_speed(camera,DC1394_ISO_SPEED_1600 );
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set ISO speed.\n");
			break;
		case 3200:
			err=dc1394_video_set_iso_speed(camera,DC1394_ISO_SPEED_3200 );
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set ISO speed.\n");
			break;
		default:
			err=dc1394_video_set_iso_speed(camera,DC1394_ISO_SPEED_MIN );
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set ISO speed.\n");
			break;

		}

		// convert the color coding to the dc1394 constants
				switch(camSettings->getCameraSetting(UAV_VIDMODE))
				{
				case UAV_GRAY:

					err=dc1394_format7_set_roi(camera,DC1394_VIDEO_MODE_FORMAT7_0,DC1394_COLOR_CODING_RAW8,DC1394_QUERY_FROM_CAMERA,
							camSettings->getCameraSetting(UAV_CCDCOL), camSettings->getCameraSetting(UAV_CCDROW), camSettings->getCameraSetting(UAV_NCOLS), camSettings->getCameraSetting(UAV_NROWS));
					DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set ROI.\n");
					break;
				case UAV_RGB8:

					err=dc1394_format7_set_roi(camera,DC1394_VIDEO_MODE_FORMAT7_0,DC1394_COLOR_CODING_RGB8,DC1394_QUERY_FROM_CAMERA,
							camSettings->getCameraSetting(UAV_CCDCOL), camSettings->getCameraSetting(UAV_CCDROW), camSettings->getCameraSetting(UAV_NCOLS), camSettings->getCameraSetting(UAV_NROWS));
					DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set ROI.\n");
					break;
				case UAV_YUV411:

					err=dc1394_format7_set_roi(camera,DC1394_VIDEO_MODE_FORMAT7_0,DC1394_COLOR_CODING_YUV411,DC1394_QUERY_FROM_CAMERA,
							camSettings->getCameraSetting(UAV_CCDCOL), camSettings->getCameraSetting(UAV_CCDROW), camSettings->getCameraSetting(UAV_NCOLS), camSettings->getCameraSetting(UAV_NROWS));
					DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set ROI.\n");
					break;
				case UAV_YUV422:

					err=dc1394_format7_set_roi(camera,DC1394_VIDEO_MODE_FORMAT7_0,DC1394_COLOR_CODING_YUV422,DC1394_QUERY_FROM_CAMERA,
							camSettings->getCameraSetting(UAV_CCDCOL), camSettings->getCameraSetting(UAV_CCDROW), camSettings->getCameraSetting(UAV_NCOLS), camSettings->getCameraSetting(UAV_NROWS));
					DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set ROI.\n");
					break;
				default: std::cerr << "Camera: ERROR unknown video mode" << std::endl; break;
				}

	// frames per second
	if(camSettings->getCameraSetting(UAV_FORMAT7) == 0)
	{
		setFps(camSettings->getCameraSetting(UAV_FPS));
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
	}

	// configure format 7
	if(camSettings->getCameraSetting(UAV_FORMAT7))
	{
		// get the max packet size
		unsigned int minPacketSize, maxPacketSize;
		err=dc1394_format7_get_packet_parameters(camera, DC1394_VIDEO_MODE_FORMAT7_0, &minPacketSize, &maxPacketSize);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR get max values for packet size.\n");
		camSettings->setCameraSetting(UAV_PACKETSIZE,maxPacketSize);

		// and we reset the packet size to obtain more frames per second
		err=dc1394_format7_set_packet_size(camera, DC1394_VIDEO_MODE_FORMAT7_0, maxPacketSize);
		DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set packet size.\n");


	}

	err=dc1394_capture_setup(camera, FIREWIRE_NUM_BUFFERS, DC1394_CAPTURE_FLAGS_DEFAULT);
	DC1394_ERR_CLN_RTN(err,shutDownCamera(), "\nCamera: ERROR capture setup.\n");

	camSettings->setCameraSetting(UAV_CAMERAUSED,1);

	return 0;
}




void CameraFirewire::startCamera( CameraSettings *camSettings )
{
	dc1394error_t err;

	// have the camera start sending us data
	err=dc1394_video_set_transmission(camera, DC1394_ON);
	//DC1394_ERR_CLN_RTN(err,shutDownCamera(), "\nCamera: ERROR start video transmission.\n");

}


void CameraFirewire::shutDownCamera( void )
{
	std::cout << "Cleaning CAMERA" << std::endl;

	dc1394_capture_stop( camera );
	dc1394_video_set_transmission( camera, DC1394_OFF );
	dc1394_iso_release_all(camera);
	dc1394_camera_set_power( camera, DC1394_OFF );


}


int CameraFirewire::init( CameraSettings *camSettings )
{
	// all the 'if(auto..)' is not necessary because if they are set to auto
	// libdc1394 will ignore the 'set_val(...)' command.

	dc1394_feature_whitebalance_set_value(camera, camSettings->getCameraSetting(UAV_WBB),camSettings->getCameraSetting(UAV_WBR));
	dc1394_feature_set_value(camera, DC1394_FEATURE_EXPOSURE, camSettings->getCameraSetting(UAV_EXPOSURE));
	dc1394_feature_set_value(camera, DC1394_FEATURE_GAIN, camSettings->getCameraSetting(UAV_GAIN));
	dc1394_feature_set_value(camera, DC1394_FEATURE_BRIGHTNESS, camSettings->getCameraSetting(UAV_BRIGHTNESS));
	dc1394_feature_set_value(camera, DC1394_FEATURE_SHUTTER, camSettings->getCameraSetting(UAV_SHUTTER));
	dc1394_feature_set_value(camera, DC1394_FEATURE_SATURATION, camSettings->getCameraSetting(UAV_SATURATION));
	dc1394_feature_set_value(camera, DC1394_FEATURE_GAMMA, camSettings->getCameraSetting(UAV_GAMMA));
	dc1394_feature_set_value(camera, DC1394_FEATURE_SHARPNESS, camSettings->getCameraSetting(UAV_SHARPNESS));

	return 0;
}

void CameraFirewire::getParameters(CameraSettings *camSettings)
{

	unsigned value, value_aux = 0;

	dc1394_feature_whitebalance_get_value(camera, &value, &value_aux);
	camSettings->setCameraSetting(UAV_WBB, value);
	camSettings->setCameraSetting(UAV_WBR, value_aux);
	dc1394_feature_get_value(camera, DC1394_FEATURE_GAIN, &value);
	camSettings->setCameraSetting(UAV_GAIN, value);
	dc1394_feature_get_value(camera, DC1394_FEATURE_SHUTTER, &value);
	camSettings->setCameraSetting(UAV_SHUTTER, value);
	dc1394_feature_get_value(camera, DC1394_FEATURE_EXPOSURE, &value);
	camSettings->setCameraSetting(UAV_EXPOSURE, value);
	dc1394_feature_get_value(camera, DC1394_FEATURE_BRIGHTNESS, &value);
	camSettings->setCameraSetting(UAV_BRIGHTNESS, value);
	dc1394_feature_get_value(camera, DC1394_FEATURE_GAMMA, &value);
	camSettings->setCameraSetting(UAV_GAMMA, value);
	dc1394_feature_get_value(camera, DC1394_FEATURE_SHARPNESS, &value);
	camSettings->setCameraSetting(UAV_SHARPNESS, value);
	dc1394_feature_get_value(camera, DC1394_FEATURE_SATURATION, &value);
	camSettings->setCameraSetting(UAV_SATURATION, value);
}


void CameraFirewire::getParametersRanges( ParameterRange parameterRange[] )
{
	dc1394_feature_get_boundaries(camera, DC1394_FEATURE_WHITE_BALANCE,
			&parameterRange[UAV_WB].min, &parameterRange[UAV_WB].max);

	dc1394_feature_get_boundaries(camera, DC1394_FEATURE_SATURATION,
			&parameterRange[UAV_SATURATION].min, &parameterRange[UAV_SATURATION].max);

	dc1394_feature_get_boundaries(camera, DC1394_FEATURE_EXPOSURE,
			&parameterRange[UAV_EXPOSURE].min, &parameterRange[UAV_EXPOSURE].max);

	dc1394_feature_get_boundaries(camera, DC1394_FEATURE_BRIGHTNESS,
			&parameterRange[UAV_BRIGHTNESS].min, &parameterRange[UAV_BRIGHTNESS].max);

	dc1394_feature_get_boundaries(camera, DC1394_FEATURE_SHUTTER,
			&parameterRange[UAV_SHUTTER].min, &parameterRange[UAV_SHUTTER].max);

	dc1394_feature_get_boundaries(camera, DC1394_FEATURE_GAIN,
			&parameterRange[UAV_GAIN].min, &parameterRange[UAV_GAIN].max);

	dc1394_feature_get_boundaries(camera, DC1394_FEATURE_GAMMA,
			&parameterRange[UAV_GAMMA].min, &parameterRange[UAV_GAMMA].max);

	dc1394_feature_get_boundaries(camera, DC1394_FEATURE_SHARPNESS,
			&parameterRange[UAV_SHARPNESS].min, &parameterRange[UAV_SHARPNESS].max);
}


int CameraFirewire::readFrame( cv::Mat &imageOpenCV )
{
	static dc1394video_frame_t *frame=NULL;
	dc1394error_t err;



	std::cerr<<"before capture dequeue!"<<std::endl;
	// ad image from camera
	err=dc1394_capture_dequeue(camera, DC1394_CAPTURE_POLICY_WAIT, &frame);

	std::cerr<<"before image creation! frame width"<<frame->size[0]<<" height: "<<frame->size[1]<<std::endl;

	imageOpenCV.create(cv::Size(frame->size[0],frame->size[1]), CV_8UC3);

	std::cerr<<"after capture dequeue!"<<std::endl;
	if (err==DC1394_SUCCESS)
	{
		std::cerr<<"before memcpy!"<<std::endl;
		memcpy(imageOpenCV.ptr(),frame->image, frame->image_bytes);
		std::cerr<<"before capture enqueue!"<<std::endl;
		dc1394_capture_enqueue (camera, frame);
	}

	return 0;
}


void CameraFirewire::setAutoManualMode(CameraSettings *camSettings )
{
	if(camSettings->getCameraSetting(UAV_AUTOBRIGHTNESS) == 1)
		dc1394_feature_set_mode(camera, DC1394_FEATURE_BRIGHTNESS, DC1394_FEATURE_MODE_AUTO);
	else
		dc1394_feature_set_mode(camera, DC1394_FEATURE_BRIGHTNESS, DC1394_FEATURE_MODE_MANUAL);

	if(camSettings->getCameraSetting(UAV_AUTOEXPOSURE) == 1)
		dc1394_feature_set_mode(camera, DC1394_FEATURE_EXPOSURE, DC1394_FEATURE_MODE_AUTO);
	else
		dc1394_feature_set_mode(camera, DC1394_FEATURE_EXPOSURE, DC1394_FEATURE_MODE_MANUAL);

	if(camSettings->getCameraSetting(UAV_AUTOGAIN) == 1)
		dc1394_feature_set_mode(camera, DC1394_FEATURE_GAIN, DC1394_FEATURE_MODE_AUTO);
	else
		dc1394_feature_set_mode(camera, DC1394_FEATURE_GAIN, DC1394_FEATURE_MODE_MANUAL);

	if(camSettings->getCameraSetting(UAV_AUTOSATURATION))
		dc1394_feature_set_mode(camera, DC1394_FEATURE_SATURATION, DC1394_FEATURE_MODE_AUTO);
	else
		dc1394_feature_set_mode(camera, DC1394_FEATURE_SATURATION, DC1394_FEATURE_MODE_MANUAL);

	if(camSettings->getCameraSetting(UAV_AUTOSHUTTER))
		dc1394_feature_set_mode(camera, DC1394_FEATURE_SHUTTER, DC1394_FEATURE_MODE_AUTO);
	else
		dc1394_feature_set_mode(camera, DC1394_FEATURE_SHUTTER, DC1394_FEATURE_MODE_MANUAL);

	if(camSettings->getCameraSetting(UAV_AUTOWB))
		dc1394_feature_set_mode(camera, DC1394_FEATURE_WHITE_BALANCE, DC1394_FEATURE_MODE_AUTO);
	else
		dc1394_feature_set_mode(camera, DC1394_FEATURE_WHITE_BALANCE, DC1394_FEATURE_MODE_MANUAL);

	dc1394_feature_set_mode(camera, DC1394_FEATURE_GAMMA, DC1394_FEATURE_MODE_MANUAL);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_SHARPNESS, DC1394_FEATURE_MODE_MANUAL);
}


void CameraFirewire::setAutoMode( CameraSettings *camSettings )
{
	dc1394_feature_set_mode(camera, DC1394_FEATURE_SHARPNESS, DC1394_FEATURE_MODE_AUTO);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_GAMMA, DC1394_FEATURE_MODE_AUTO);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_BRIGHTNESS, DC1394_FEATURE_MODE_AUTO);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_EXPOSURE, DC1394_FEATURE_MODE_AUTO);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_GAIN, DC1394_FEATURE_MODE_AUTO);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_SATURATION, DC1394_FEATURE_MODE_AUTO);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_SHUTTER, DC1394_FEATURE_MODE_AUTO);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_WHITE_BALANCE, DC1394_FEATURE_MODE_AUTO);
}


void CameraFirewire::setManualMode( CameraSettings *camSettings )
{
	dc1394_feature_set_power(camera, DC1394_FEATURE_SHARPNESS, DC1394_ON);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_SHARPNESS, DC1394_FEATURE_MODE_MANUAL);
	dc1394_feature_set_power(camera, DC1394_FEATURE_GAMMA,DC1394_ON);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_GAMMA, DC1394_FEATURE_MODE_MANUAL);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_BRIGHTNESS, DC1394_FEATURE_MODE_MANUAL);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_EXPOSURE, DC1394_FEATURE_MODE_MANUAL);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_GAIN, DC1394_FEATURE_MODE_MANUAL);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_SATURATION, DC1394_FEATURE_MODE_MANUAL);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_SHUTTER, DC1394_FEATURE_MODE_MANUAL);
	dc1394_feature_set_mode(camera, DC1394_FEATURE_WHITE_BALANCE, DC1394_FEATURE_MODE_MANUAL);
}


void CameraFirewire::printParameters( void )
{
	std::cerr << "Camera: dc1394 settings" << std::endl;

	unsigned int value, value_aux = 0;

	dc1394_feature_whitebalance_get_value(camera, &value, &value_aux);
	std::cerr<<"WBB: "<<value<<"WBR: "<<value_aux<<std::endl;

	dc1394_feature_get_value(camera, DC1394_FEATURE_GAIN, &value);
	std::cerr<<"Gain: "<<value<<std::endl;

	dc1394_feature_get_value(camera, DC1394_FEATURE_SHUTTER, &value);
	std::cerr<<"Shutter: "<<value<<std::endl;

	dc1394_feature_get_value(camera, DC1394_FEATURE_EXPOSURE, &value);
	std::cerr<<"Exposure: "<<value<<std::endl;

	dc1394_feature_get_value(camera, DC1394_FEATURE_BRIGHTNESS, &value);
	std::cerr<<"Brightness: "<<value<<std::endl;

	dc1394_feature_get_value(camera, DC1394_FEATURE_GAMMA, &value);
	std::cerr<<"Gamma: "<<value<<std::endl;

	dc1394_feature_get_value(camera, DC1394_FEATURE_SHARPNESS, &value);
	std::cerr<<"Sharpness: "<<value<<std::endl;

	dc1394_feature_get_value(camera, DC1394_FEATURE_SATURATION, &value);
	std::cerr<<"Saturation: "<<value<<std::endl;


	dc1394_feature_get_value(camera, DC1394_FEATURE_FRAME_RATE, &value);
	std::cerr<<"Fps: "<<value<<std::endl;
}

void CameraFirewire::printParameter(unsigned int parameter)
{
	unsigned int value, value_aux = 0;
	switch(parameter)
	{
	case UAV_FPS:
		dc1394_feature_get_value(camera, DC1394_FEATURE_FRAME_RATE, &value);
		std::cerr<<"Fps: "<<value<<std::endl;
		break;
	case UAV_SATURATION:
		dc1394_feature_get_value(camera, DC1394_FEATURE_SATURATION, &value);
		std::cerr<<"Saturation: "<<value<<std::endl;
		break;
	case UAV_WB:
		dc1394_feature_whitebalance_get_value(camera, &value, &value_aux);
		std::cerr<<"WBB: "<<value<<"WBR: "<<value_aux<<std::endl;
		break;

	case UAV_GAIN:
		dc1394_feature_get_value(camera, DC1394_FEATURE_GAIN, &value);
		std::cerr<<"Gain: "<<value<<std::endl;
		break;

	case UAV_SHUTTER:

		dc1394_feature_get_value(camera, DC1394_FEATURE_SHUTTER, &value);
		std::cerr<<"Shutter: "<<value<<std::endl;
		break;

	case UAV_EXPOSURE:

		dc1394_feature_get_value(camera, DC1394_FEATURE_EXPOSURE, &value);
		std::cerr<<"Exposure: "<<value<<std::endl;
		break;

	case UAV_BRIGHTNESS:
		dc1394_feature_get_value(camera, DC1394_FEATURE_BRIGHTNESS, &value);
		std::cerr<<"Brightness: "<<value<<std::endl;
		break;

	case UAV_GAMMA:
		dc1394_feature_get_value(camera, DC1394_FEATURE_GAMMA, &value);
		std::cerr<<"Gamma: "<<value<<std::endl;
		break;

	case UAV_SHARPNESS:
		dc1394_feature_get_value(camera, DC1394_FEATURE_SHARPNESS, &value);
		std::cerr<<"Sharpness: "<<value<<std::endl;
		break;

	default:
		std::cerr<<"Parameter not supported!"<<std::endl;
		break;

	}
}


	void CameraFirewire:: stopCamera(CameraSettings *camSettings )
	{
		std::cerr<<"Method not supported!"<<std::endl;
	}

	void CameraFirewire::setParameters(CameraSettings *camSettings)
	{
		setFps(camSettings->getCameraSetting(UAV_FPS));

		dc1394_feature_whitebalance_set_value(camera, camSettings->getCameraSetting(UAV_WBB),camSettings->getCameraSetting(UAV_WBR));
		dc1394_feature_set_value(camera, DC1394_FEATURE_EXPOSURE, camSettings->getCameraSetting(UAV_EXPOSURE));
		dc1394_feature_set_value(camera, DC1394_FEATURE_GAIN, camSettings->getCameraSetting(UAV_GAIN));
		dc1394_feature_set_value(camera, DC1394_FEATURE_BRIGHTNESS, camSettings->getCameraSetting(UAV_BRIGHTNESS));
		dc1394_feature_set_value(camera, DC1394_FEATURE_SHUTTER, camSettings->getCameraSetting(UAV_SHUTTER));
		dc1394_feature_set_value(camera, DC1394_FEATURE_SATURATION, camSettings->getCameraSetting(UAV_SATURATION));
		dc1394_feature_set_value(camera, DC1394_FEATURE_GAMMA, camSettings->getCameraSetting(UAV_GAMMA));
		dc1394_feature_set_value(camera, DC1394_FEATURE_SHARPNESS, camSettings->getCameraSetting(UAV_SHARPNESS));


	}

	void CameraFirewire::setParameter(unsigned int parameter, unsigned value)
	{
		switch(parameter)
		{
		case UAV_FPS:
			setFps(value);
			break;
		case UAV_WB:
			dc1394_feature_whitebalance_set_value(camera, value, value); //FIXME
			break;
		case UAV_EXPOSURE:
			dc1394_feature_set_value(camera, DC1394_FEATURE_EXPOSURE, value);
			break;
		case UAV_GAIN:
			dc1394_feature_set_value(camera, DC1394_FEATURE_GAIN, value);
			break;
		case UAV_BRIGHTNESS:
			dc1394_feature_set_value(camera, DC1394_FEATURE_BRIGHTNESS, value);
			break;
		case UAV_SHUTTER:
			dc1394_feature_set_value(camera, DC1394_FEATURE_SHUTTER, value);
			break;
		case UAV_SATURATION:
			dc1394_feature_set_value(camera, DC1394_FEATURE_SATURATION, value);
			break;
		case UAV_GAMMA:
			dc1394_feature_set_value(camera, DC1394_FEATURE_GAMMA, value);
			break;
		case UAV_SHARPNESS:
			dc1394_feature_set_value(camera, DC1394_FEATURE_SHARPNESS, value);
			break;
		default:
			std::cerr<<"Parameter not supported!"<<std::endl;
			break;
		}

	}

	void CameraFirewire::getParameter(unsigned int parameter, double *valueDouble)
	{
		unsigned int value_aux, value;
		value = unsigned(*valueDouble);

		switch(parameter)
		{
		case UAV_WB:
			dc1394_feature_whitebalance_get_value(camera, &value, &value_aux);
			break;
		case UAV_GAIN:
			dc1394_feature_get_value(camera, DC1394_FEATURE_GAIN, &value);
			break;
		case UAV_SHUTTER:
			dc1394_feature_get_value(camera, DC1394_FEATURE_SHUTTER, &value);
			break;
		case UAV_EXPOSURE:
			dc1394_feature_get_value(camera, DC1394_FEATURE_EXPOSURE, &value);
			break;
		case UAV_BRIGHTNESS:
			dc1394_feature_get_value(camera, DC1394_FEATURE_BRIGHTNESS, &value);
			break;
		case UAV_GAMMA:
			dc1394_feature_get_value(camera, DC1394_FEATURE_GAMMA, &value);
			break;
		case UAV_SHARPNESS:
			dc1394_feature_get_value(camera, DC1394_FEATURE_SHARPNESS, &value);
			break;
		case UAV_SATURATION:
			dc1394_feature_get_value(camera, DC1394_FEATURE_SATURATION, &value);
			break;
		default:
			std::cerr<<"Parameter not supported!"<<std::endl;
			break;
		}

		*valueDouble = value;
	}

	void CameraFirewire::getParameterRanges(unsigned int parameter,ParameterRange parameterRange)
	{

		switch(parameter)
		{
		case UAV_WB:
			dc1394_feature_get_boundaries(camera, DC1394_FEATURE_WHITE_BALANCE,
					&parameterRange.min, &parameterRange.max);
			break;
		case UAV_SATURATION:
			dc1394_feature_get_boundaries(camera, DC1394_FEATURE_SATURATION,
					&parameterRange.min, &parameterRange.max);
			break;
		case UAV_EXPOSURE:
			dc1394_feature_get_boundaries(camera, DC1394_FEATURE_EXPOSURE,
					&parameterRange.min, &parameterRange.max);
			break;
		case UAV_BRIGHTNESS:
			dc1394_feature_get_boundaries(camera, DC1394_FEATURE_BRIGHTNESS,
					&parameterRange.min, &parameterRange.max);
			break;
		case UAV_SHUTTER:
			dc1394_feature_get_boundaries(camera, DC1394_FEATURE_SHUTTER,
					&parameterRange.min, &parameterRange.max);
			break;
		case UAV_GAIN:
			dc1394_feature_get_boundaries(camera, DC1394_FEATURE_GAIN,
					&parameterRange.min, &parameterRange.max);
			break;
		case UAV_GAMMA:
			dc1394_feature_get_boundaries(camera, DC1394_FEATURE_GAMMA,
					&parameterRange.min, &parameterRange.max);
			break;
		case UAV_SHARPNESS:
			dc1394_feature_get_boundaries(camera, DC1394_FEATURE_SHARPNESS,
					&parameterRange.min, &parameterRange.max);
			break;
		default:
			std::cerr<<"Parameter not supported!"<<std::endl;
			break;

		}
	}

int CameraFirewire::setFps(unsigned int fps)
{
	dc1394error_t err;

	switch(fps)
		{
		case 1:
			err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_1_875);
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
			break;
		case 3:
			err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_3_75);
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
			break;
		case 7:
			err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_7_5);
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
			break;
		case 15:
			err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_15);
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
			break;
		case 30:
			err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_30);
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
			break;

		case 60:
			err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_60);
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
			break;
		case 120:
			err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_120);
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
			break;
		case 240:
			err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_240);
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
			break;

		default:
			err=dc1394_video_set_framerate(camera, DC1394_FRAMERATE_MIN);
			DC1394_ERR_CLN_RTN(err,shutDownCamera(),"\nCamera: ERROR set frame rate.\n");
			break;
		}
	return 0;
}
}
